import pytest

from flex.serializers.definitions import BaseSchemaSerializer
from flex.constants import (
    NULL,
    BOOLEAN,
    INTEGER,
    NUMBER,
    STRING,
    ARRAY,
    OBJECT,
)

from tests.utils import assert_error_message_equal


#
# minimum validation tests
#
@pytest.mark.parametrize(
    'type_',
    (NULL, BOOLEAN, STRING, ARRAY, OBJECT),
)
def test_minimum_is_invalid_for_non_numeric_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'minimum': 0,
        },
    )
    assert not serializer.is_valid()
    assert 'minimum' in serializer.errors
    assert_error_message_equal(
        serializer.errors['minimum'][0],
        serializer.error_messages['invalid_type_for_minimum'],
    )


@pytest.mark.parametrize(
    'type_',
    (INTEGER, NUMBER),
)
def test_minimum_is_valid_with_numeric_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'minimum': 0,
        },
    )
    assert 'minimum' not in serializer.errors


@pytest.mark.parametrize(
    'types',
    (
        (INTEGER, NUMBER),
        (STRING, NUMBER),
        (NUMBER, NULL),
    ),
)
def test_minimum_is_valid_if_at_least_one_type_is_numeric(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'minimum': 0,
        },
    )
    assert 'minimum' not in serializer.errors


#
# maximum validation tests
#
@pytest.mark.parametrize(
    'type_',
    (NULL, BOOLEAN, STRING, ARRAY, OBJECT),
)
def test_maximum_is_invalid_for_non_numeric_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'maximum': 0,
        },
    )
    assert not serializer.is_valid()
    assert 'maximum' in serializer.errors
    assert_error_message_equal(
        serializer.errors['maximum'][0],
        serializer.error_messages['invalid_type_for_maximum'],
    )


@pytest.mark.parametrize(
    'type_',
    (INTEGER, NUMBER),
)
def test_maximum_is_valid_with_numeric_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'maximum': 0,
        },
    )
    assert 'maximum' not in serializer.errors


@pytest.mark.parametrize(
    'types',
    (
        (INTEGER, NUMBER),
        (STRING, NUMBER),
        (NUMBER, NULL),
    ),
)
def test_maximum_is_valid_if_at_least_one_type_is_numeric(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'maximum': 0,
        },
    )
    assert 'maximum' not in serializer.errors


#
# exclusiveMinimum tests
#
def test_exclusive_minimum_requires_minimum():
    serializer = BaseSchemaSerializer(
        data={
            #'maximum': 0,  # Intentionally commented out to show it's missing from data.
            'exclusiveMinimum': True
        },
    )
    assert not serializer.is_valid()
    assert 'exclusiveMinimum' in serializer.errors
    assert_error_message_equal(
        serializer.errors['exclusiveMinimum'][0],
        serializer.error_messages['exclusive_minimum_requires_minimum'],
    )


def test_exclusive_minimum_with_minimum_is_valid():
    serializer = BaseSchemaSerializer(
        data={
            'minimum': 0,
            'exclusiveMinimum': True
        },
    )
    assert 'exclusiveMinimum' not in serializer.errors


#
# exclusiveMaximum tests
#
def test_exclusive_maximum_requires_maximum():
    serializer = BaseSchemaSerializer(
        data={
            #'maximum': 0,  # Intentionally commented out to show it's missing from data.
            'exclusiveMaximum': True
        },
    )
    assert not serializer.is_valid()
    assert 'exclusiveMaximum' in serializer.errors
    assert_error_message_equal(
        serializer.errors['exclusiveMaximum'][0],
        serializer.error_messages['exclusive_maximum_requires_maximum'],
    )


def test_exclusive_maximum_with_maximum_is_valid():
    serializer = BaseSchemaSerializer(
        data={
            'maximum': 0,
            'exclusiveMaximum': True
        },
    )
    assert 'exclusiveMaximum' not in serializer.errors


#
# multipleOf validation tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        STRING,
        ARRAY,
        OBJECT,
    ),
)
def test_multiple_of_invalid_with_non_number_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'multipleOf': 10,
        },
    )
    assert not serializer.is_valid()
    assert 'multipleOf' in serializer.errors
    assert_error_message_equal(
        serializer.errors['multipleOf'][0],
        serializer.error_messages['invalid_type_for_multiple_of'],
    )


@pytest.mark.parametrize(
    'type_',
    (INTEGER, NUMBER),
)
def test_multiple_of_valid_for_json_number_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'multipleOf': 10,
        },
    )
    assert 'multipleOf' not in serializer.errors


@pytest.mark.parametrize(
    'types',
    (
        (INTEGER, NUMBER),
        (STRING, NUMBER),
        (INTEGER, NULL),
    )
)
def test_multiple_of_valid_if_any_of_declared_types_are_numeric(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'multipleOf': 10,
        },
    )
    assert 'multipleOf' not in serializer.errors


#
# minLength validation tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        ARRAY,
        OBJECT,
    ),
)
def test_min_length_invalid_with_non_string_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'minLength': 10,
        },
    )
    assert not serializer.is_valid()
    assert 'minLength' in serializer.errors
    assert_error_message_equal(
        serializer.errors['minLength'][0],
        serializer.error_messages['invalid_type_for_min_length'],
    )


def test_min_length_valid_with_string_type():
    serializer = BaseSchemaSerializer(
        data={
            'type': STRING,
            'minLength': 10,
        },
    )
    assert 'minLength' not in serializer.errors


@pytest.mark.parametrize(
    'types',
    (
        (STRING,),
        (NULL, STRING),
        (STRING, INTEGER),
    ),
)
def test_min_length_valid_if_any_declared_type_is_string(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'minLength': 10,
        },
    )
    assert 'minLength' not in serializer.errors


#
# maxLength validation tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        ARRAY,
        OBJECT,
    ),
)
def test_max_length_invalid_with_non_string_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'maxLength': 10,
        },
    )
    assert not serializer.is_valid()
    assert 'maxLength' in serializer.errors
    assert_error_message_equal(
        serializer.errors['maxLength'][0],
        serializer.error_messages['invalid_type_for_max_length'],
    )


def test_max_length_valid_with_string_type():
    serializer = BaseSchemaSerializer(
        data={
            'type': STRING,
            'maxLength': 10,
        },
    )
    assert 'maxLength' not in serializer.errors


@pytest.mark.parametrize(
    'types',
    (
        (STRING,),
        (NULL, STRING),
        (STRING, INTEGER),
    ),
)
def test_max_length_valid_if_any_declared_type_is_string(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'maxLength': 10,
        },
    )
    assert 'maxLength' not in serializer.errors


#
# maxItems tests
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        OBJECT,
    ),
)
def test_max_items_invalid_with_non_array_type(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'maxItems': 5,
        },
    )
    assert not serializer.is_valid()
    assert 'maxItems' in serializer.errors
    assert_error_message_equal(
        serializer.errors['maxItems'][0],
        serializer.error_messages['invalid_type_for_max_items'],
    )


@pytest.mark.parametrize(
    'types',
    (
        ARRAY,
        (ARRAY,),
        (STRING, ARRAY),
        (ARRAY, OBJECT),
    ),
)
def test_max_items_valid_with_array_type(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'maxItems': 5,
        },
    )
    assert 'maxItems' not in serializer.errors


#
# minItems tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        OBJECT,
    ),
)
def test_min_items_invalid_with_non_arraw_type(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'minItems': 5,
        },
    )
    assert not serializer.is_valid()
    assert 'minItems' in serializer.errors
    assert_error_message_equal(
        serializer.errors['minItems'][0],
        serializer.error_messages['invalid_type_for_min_items'],
    )


@pytest.mark.parametrize(
    'types',
    (
        ARRAY,
        (ARRAY,),
        (STRING, ARRAY),
        (ARRAY, OBJECT),
    ),
)
def test_min_items_valid_with_array_type(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'minItems': 5,
        },
    )
    assert 'minItems' not in serializer.errors


#
# uniqueItems tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        OBJECT,
    ),
)
def test_unique_items_invalid_with_non_arraw_type(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'uniqueItems': True,
        },
    )
    assert not serializer.is_valid()
    assert 'uniqueItems' in serializer.errors
    assert_error_message_equal(
        serializer.errors['uniqueItems'][0],
        serializer.error_messages['invalid_type_for_unique_items'],
    )


@pytest.mark.parametrize(
    'types',
    (
        ARRAY,
        (BOOLEAN, ARRAY),
        (INTEGER, ARRAY),
        (ARRAY, OBJECT),
    ),
)
def test_unique_items_valid_with_array_type(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'uniqueItems': True,
        },
    )
    assert 'uniqueItems' not in serializer.errors


#
# enum tests
#
@pytest.mark.parametrize(
    'enum',
    (
        'abcd',
        {'a': 'foo', 'b': 'bar'},
        123,
        None,
    ),
)
def test_non_array_for_enum_is_invalid(enum):

    serializer = BaseSchemaSerializer(
        data={
            'enum': enum,
        },
    )
    assert not serializer.is_valid()
    assert 'enum' in serializer.errors


def test_valid_enum_iterable():
    serializer = BaseSchemaSerializer(
        data={
            'enum': ['a', 1, True, ['inner', 'array'], {'a': 'b', 'c': 'd'}, 2.0, None],
        },
    )
    assert 'enum' not in serializer.errors


#
# minProperties tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        ARRAY,
    ),
)
def test_min_properties_invalid_for_non_object_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'minProperties': 3,
        },
    )
    assert not serializer.is_valid()
    assert 'minProperties' in serializer.errors
    assert_error_message_equal(
        serializer.errors['minProperties'][0],
        serializer.error_messages['invalid_type_for_min_properties']
    )


@pytest.mark.parametrize(
    'types',
    (
        OBJECT,
        (OBJECT, ARRAY),
        (OBJECT, NULL),
        (INTEGER, STRING, OBJECT),
    ),
)
def test_min_properties_valid_for_object_type(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'minProperties': 3,
        },
    )
    assert 'minProperties' not in serializer.errors


def test_min_properties_must_be_strictly_positive():
    serializer = BaseSchemaSerializer(
        data={
            'type': OBJECT,
            'minProperties': -1,
        },
    )
    assert not serializer.is_valid()
    assert 'minProperties' in serializer.errors
    assert serializer.errors['minProperties'][0].startswith(
        "Ensure this value is greater than or equal to",
    )


#
# maxProperties tests
#
@pytest.mark.parametrize(
    'type_',
    (
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        ARRAY,
    ),
)
def test_max_properties_invalid_for_non_object_types(type_):
    serializer = BaseSchemaSerializer(
        data={
            'type': type_,
            'maxProperties': 3,
        },
    )
    assert not serializer.is_valid()
    assert 'maxProperties' in serializer.errors
    assert_error_message_equal(
        serializer.errors['maxProperties'][0],
        serializer.error_messages['invalid_type_for_max_properties']
    )


@pytest.mark.parametrize(
    'types',
    (
        OBJECT,
        (OBJECT, ARRAY),
        (OBJECT, NULL),
        (INTEGER, STRING, OBJECT),
    ),
)
def test_max_properties_valid_for_object_type(types):
    serializer = BaseSchemaSerializer(
        data={
            'type': types,
            'maxProperties': 3,
        },
    )
    assert 'maxProperties' not in serializer.errors


def test_max_properties_must_be_strictly_positive():
    serializer = BaseSchemaSerializer(
        data={
            'type': OBJECT,
            'maxProperties': -1,
        },
    )
    assert not serializer.is_valid()
    assert 'maxProperties' in serializer.errors
    assert serializer.errors['maxProperties'][0].startswith(
        "Ensure this value is greater than or equal to",
    )



import re
import decimal
import operator
import functools
import collections
import itertools

import six

from django.core.exceptions import ValidationError
from django.core.validators import (
    MinLengthValidator,
    MaxLengthValidator,
)

from flex.context_managers import ErrorCollection
from flex.formats import registry
from flex.utils import (
    is_value_of_any_type,
    is_non_string_iterable,
    get_type_for_value,
    chain_reduce_partial,
    cast_value_to_type,
)
from flex.constants import (
    EMPTY,
    NUMBER,
    STRING,
    ARRAY,
    DELIMETERS,
)
from flex.decorators import (
    skip_if_not_of_type,
    suffix_reserved_words,
)
from flex.error_messages import MESSAGES


def skip_if_empty(func):
    """
    Decorator for validation functions which makes them pass if the value
    passed in is the EMPTY sentinal value.
    """
    @functools.wraps(func)
    def inner(value, *args, **kwargs):
        if value is EMPTY:
            return
        else:
            return func(value, *args, **kwargs)
    return inner


@skip_if_empty
def validate_type(value, types):
    """
    Validate that the value is one of the provided primative types.
    """
    if not is_value_of_any_type(value, types):
        raise ValidationError(MESSAGES['type']['invalid'].format(
            repr(value), get_type_for_value(value), types,
        ))


@suffix_reserved_words
def generate_type_validator(type_, **kwargs):
    """
    Generates a callable validator for the given type or iterable of types.
    """
    if is_non_string_iterable(type_):
        types = type_
    else:
        types = (type_,)
    return functools.partial(validate_type, types=types)


@suffix_reserved_words
def generate_format_validator(format_, **kwargs):
    """
    Returns the validator function for the given format, as registered with the
    format registry.
    """
    if format_ in registry:
        return registry[format_]
    else:
        return noop


def noop(*args, **kwargs):
    """
    No-Op validator that does nothing.
    """
    pass


def validate_required(value):
    if value is EMPTY:
        raise ValidationError(MESSAGES['required']['required'])


def generate_required_validator(required, **kwargs):
    if required:
        return validate_required
    else:
        return noop


@skip_if_empty
@skip_if_not_of_type(NUMBER)
def validate_multiple_of(value, divisor):
    """
    Given a value and a divisor, validate that the value is divisible by the
    divisor.
    """
    if not decimal.Decimal(str(value)) % decimal.Decimal(str(divisor)) == 0:
        raise ValidationError(
            MESSAGES['multiple_of']['invalid'].format(divisor, value),
        )


def generate_multiple_of_validator(multipleOf, **kwargs):
    return functools.partial(validate_multiple_of, divisor=multipleOf)


@skip_if_empty
@skip_if_not_of_type(NUMBER)
def validate_minimum(value, minimum, is_exclusive):
    """
    Validator function for validating that a value does not violate it's
    minimum allowed value.  This validation can be inclusive, or exclusive of
    the minimum depending on the value of `is_exclusive`.
    """
    if is_exclusive:
        comparison_text = "greater than"
        compare_fn = operator.gt
    else:
        comparison_text = "greater than or equal to"
        compare_fn = operator.ge

    if not compare_fn(value, minimum):
        raise ValidationError(
            MESSAGES['minimum']['invalid'].format(value, comparison_text, minimum),
        )


def generate_minimum_validator(minimum, exclusiveMinimum=False, **kwargs):
    """
    Generator function returning a callable for minimum value validation.
    """
    return functools.partial(validate_minimum, minimum=minimum, is_exclusive=exclusiveMinimum)


@skip_if_empty
@skip_if_not_of_type(NUMBER)
def validate_maximum(value, maximum, is_exclusive):
    """
    Validator function for validating that a value does not violate it's
    maximum allowed value.  This validation can be inclusive, or exclusive of
    the maximum depending on the value of `is_exclusive`.
    """
    if is_exclusive:
        comparison_text = "less than"
        compare_fn = operator.lt
    else:
        comparison_text = "less than or equal to"
        compare_fn = operator.le

    if not compare_fn(value, maximum):
        raise ValidationError(
            MESSAGES['maximum']['invalid'].format(value, comparison_text, maximum),
        )


def generate_maximum_validator(maximum, exclusiveMaximum=False, **kwargs):
    """
    Generator function returning a callable for maximum value validation.
    """
    return functools.partial(validate_maximum, maximum=maximum, is_exclusive=exclusiveMaximum)


def generate_min_length_validator(minLength, **kwargs):
    """
    Generates a validator for enforcing the minLength of a string.
    """
    return skip_if_empty(skip_if_not_of_type(STRING)(MinLengthValidator(minLength).__call__))


def generate_max_length_validator(maxLength, **kwargs):
    """
    Generates a validator for enforcing the maxLength of a string.
    """
    return skip_if_empty(skip_if_not_of_type(STRING)(MaxLengthValidator(maxLength).__call__))


@skip_if_empty
@skip_if_not_of_type(ARRAY)
def validate_min_items(value, minimum):
    """
    Validator for ARRAY types to enforce a minimum number of items allowed for
    the ARRAY to be valid.
    """
    if len(value) < minimum:
        raise ValidationError(
            MESSAGES['min_items']['invalid'].format(
                minimum, len(value),
            ),
        )


def generate_min_items_validator(minItems, **kwargs):
    """
    Generator function returning a callable for minItems validation.
    """
    return functools.partial(validate_min_items, minimum=minItems)


@skip_if_empty
@skip_if_not_of_type(ARRAY)
def validate_max_items(value, maximum):
    """
    Validator for ARRAY types to enforce a maximum number of items allowed for
    the ARRAY to be valid.
    """
    if len(value) > maximum:
        raise ValidationError(
            MESSAGES['max_items']['invalid'].format(
                maximum, len(value),
            ),
        )


def generate_max_items_validator(maxItems, **kwargs):
    """
    Generator function returning a callable for maxItems validation.
    """
    return functools.partial(validate_max_items, maximum=maxItems)


@skip_if_empty
@skip_if_not_of_type(ARRAY)
def validate_unique_items(value):
    """
    Validator for ARRAY types to enforce that all array items must be unique.
    """
    # we can't just look at the items themselves since 0 and False are treated
    # the same as dictionary keys.
    counter = collections.Counter((
        (v, type(v)) for v in value
    ))
    dupes = [v[0] for v, count in counter.items() if count > 1]
    if dupes:
        raise ValidationError(
            MESSAGES['unique_items']['invalid'].format(
                repr(dupes),
            ),
        )


def generate_unique_items_validator(uniqueItems, **kwargs):
    """
    Returns the unique_item_validator if uniqueItems is set to True, otherwise
    it returns the noop function.
    """
    if uniqueItems:
        return validate_unique_items
    else:
        return noop


@skip_if_empty
@skip_if_not_of_type(STRING)
def validate_pattern(value, regex):
    if not regex.match(value):
        raise ValidationError(
            MESSAGES['pattern']['invalid'].format(value, regex.pattern),
        )


def generate_pattern_validator(pattern, **kwargs):
    return functools.partial(validate_pattern, regex=re.compile(pattern))


def deep_equal(a, b):
    """
    Because of things in python like:
        >>> 1 == 1.0
        True
        >>> 1 == True
        True
    """
    return a == b and isinstance(a, type(b)) and isinstance(b, type(a))


@skip_if_empty
def validate_enum(value, options):
    if not any(deep_equal(value, option) for option in options):
        raise ValidationError(
            MESSAGES['enum']['invalid'].format(
                value, options,
            )
        )


def generate_enum_validator(enum, **kwargs):
    return functools.partial(validate_enum, options=enum)


def validate_object(obj, validators, inner=False):
    """
    Takes a mapping and applies a mapping of validator functions to it
    collecting and reraising any validation errors that occur.
    """
    with ErrorCollection(inner=inner) as errors:
        if '$ref' in validators:
            ref_ = validators.pop('$ref')
            for k, v in ref_.validators.items():
                validators.setdefault(k, v)
        for key, validator in validators.items():
            try:
                validator(obj)
            except ValidationError as err:
                errors[key].extend(list(err.messages))


@suffix_reserved_words
def generate_value_processor(type_, collectionFormat=None, items=None, **kwargs):
    """
    Create a callable that will take the string value of a header and cast it
    to the appropriate type.  This can involve:

    - splitting a header of type 'array' by its delimeters.
    - type casting the internal elements of the array.
    """
    processors = []
    if is_non_string_iterable(type_):
        assert False, "This should not be possible"
    else:
        if type_ == ARRAY and collectionFormat:
            delimeter = DELIMETERS[collectionFormat]
            # split the string based on the delimeter specified by the
            # `collectionFormat`
            processors.append(operator.methodcaller('split', delimeter))
            # remove any Falsy values like empty strings.
            processors.append(functools.partial(filter, bool))
            # strip off any whitespace
            processors.append(functools.partial(map, operator.methodcaller('strip')))
            if items is not None:
                if isinstance(items, collections.Mapping):
                    items_processors = itertools.repeat(
                        generate_value_processor(**items)
                    )
                elif isinstance(items, collections.Sequence):
                    items_processors = itertools.chain(
                        (generate_value_processor(**item) for item in items),
                        itertools.repeat(lambda v: v),
                    )
                elif isinstance(items, six.string_types):
                    raise NotImplementedError("Not implemented")
                else:
                    assert False, "Should not be possible"
                # 1. zip the processor and the array items together
                # 2. apply the processor to each array item.
                # 3. cast the starmap generator to a list.
                processors.append(
                    chain_reduce_partial(
                        functools.partial(zip, items_processors),
                        functools.partial(itertools.starmap, lambda fn, v: fn(v)),
                        list,
                    )
                )
        else:
            processors.append(
                functools.partial(cast_value_to_type, type_=type_)
            )

    def processor(value):
        try:
            return chain_reduce_partial(*processors)(value)
        except (ValueError, TypeError):
            return value

    return processor



import click

from flex.core import load


@click.command()
@click.option(
    '-s', '--source',
    help='Source; a url to a schema or a file path to a schema',
)
def main(source):
    """
    For a given command line supplied argument, negotiate the content, parse
    the schema and then return any issues to stdout or if no schema issues,
    return success exit code.
    """
    if source is None:
        click.echo(
            "You need to supply a file or url to a schema to a swagger schema, for"
            "the validator to work."
        )
        return 1
    try:
        load(source)
        click.echo("Validation passed")
        return 0
    except ValueError, e:
        raise click.ClickException(e.message)



import urllib
import urllib2
import urlparse
import json

from flex.constants import EMPTY


class URLMixin(object):
    @property
    def url_components(self):
        return urlparse.urlparse(self.url)

    @property
    def path(self):
        return self.url_components.path

    @property
    def query(self):
        return self.url_components.query

    @property
    def query_data(self):
        return urlparse.parse_qs(self.query)


class Request(URLMixin):
    """
    Generic request object.  All supported requests are normalized to an
    instance of Request.
    """
    method = None

    def __init__(self, url, method, content_type=None, body=None, request=None, headers=None):
        self._request = request
        self.body = body
        self.url = url
        self.method = method
        self.content_type = content_type
        self.headers = headers or {}


def _normalize_requests_request(request):
    import requests

    if not isinstance(request, (requests.Request, requests.PreparedRequest)):
        raise TypeError("Cannot normalize this request object")

    url = request.url
    method = request.method.lower()
    content_type = request.headers.get('Content-Type')
    body = request.body

    return Request(
        url=url,
        body=body,
        method=method,
        content_type=content_type,
        request=request,
    )


def _normalize_urllib2_request(request):
    if not isinstance(request, urllib2.Request):
        raise TypeError("Cannot normalize this request object")

    url = request.get_full_url()
    method = request.get_method().lower()
    content_type = request.headers.get('Content-type')
    body = request.get_data()

    return Request(
        url=url,
        body=body,
        method=method,
        content_type=content_type,
        request=request,
    )


REQUEST_NORMALIZERS = (
    _normalize_urllib2_request,
    _normalize_requests_request,
)


def normalize_request(request):
    """
    Given a request, normalize it to the internal Request class.
    """
    if isinstance(request, Request):
        return request

    for normalizer in REQUEST_NORMALIZERS:
        try:
            return normalizer(request)
        except TypeError:
            continue

    raise ValueError("Unable to normalize the provided request")


class Response(URLMixin):
    """
    Generic response object.  All supported responses are normalized to an
    instance of this Response.
    """
    _response = None
    status_code = None

    def __init__(self, request, content, url, status_code, content_type,
                 headers=None, response=None):
        self._response = response
        self.request = request
        self.content = content
        self.url = url
        self.status_code = status_code
        self.content_type = content_type
        self.headers = headers or {}

    @property
    def path(self):
        return urlparse.urlparse(self.url).path

    @property
    def data(self):
        if self.content is EMPTY:
            return self.content
        elif self.content_type == 'application/json':
            return json.loads(self.content)
        raise NotImplementedError("No content negotiation for this content type")


def _normalize_requests_response(response, request=None):
    import requests
    if not isinstance(response, requests.Response):
        raise TypeError("Cannot normalize this response object")

    url = response.url
    status_code = response.status_code
    content_type = response.headers.get('Content-Type')

    return Response(
        request=request,
        content=response.content,
        url=url,
        status_code=status_code,
        content_type=content_type,
        response=response,
    )


def _normalize_urllib_response(response, request=None):
    if not isinstance(response, urllib.addinfourl):
        raise TypeError("Cannot normalize this response object")

    url = response.url
    status_code = response.getcode()
    content_type = response.headers.get('Content-Type')

    return Response(
        request=request,
        content=response.read(),
        url=url,
        status_code=status_code,
        content_type=content_type,
        response=response,
    )


RESPONSE_NORMALIZERS = (
    _normalize_urllib_response,
    _normalize_requests_response,
)


def normalize_response(response, request=None):
    """
    Given a response, normalize it to the internal Response class.  This also
    involves normalizing the associated request object.
    """
    if isinstance(response, Response):
        return response
    if request is not None and not isinstance(request, Request):
        request = normalize_request(request)

    for normalizer in RESPONSE_NORMALIZERS:
        try:
            return normalizer(response, request=request)
        except TypeError:
            continue

    raise ValueError("Unable to normalize the provided response")



import pytest

from flex.serializers.core import (
    HeaderSerializer,
)
from flex.constants import (
    BOOLEAN,
    NUMBER,
    ARRAY,
    STRING,
)

from tests.utils import assert_error_message_equal


def test_items_is_required_when_type_is_array():
    serializer = HeaderSerializer(
        data={'type': ARRAY},
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert_error_message_equal(
        serializer.errors['items'][0],
        serializer.error_messages['items_required'],
    )


@pytest.mark.parametrize(
    'type_,default',
    (
        (BOOLEAN, 'abc'),
        (BOOLEAN, 1),
        (NUMBER, True),
        (NUMBER, 'abc'),
        (ARRAY, 'abc'),
        (STRING, 1),
        (STRING, None),
    )
)
def test_mistyped_header_default_boolean_to_string(type_, default):
    serializer = HeaderSerializer(
        data={'type': type_, 'default': default}
    )

    assert not serializer.is_valid()
    assert 'default' in serializer.errors
    assert_error_message_equal(
        serializer.errors['default'][0],
        serializer.error_messages['default_is_incorrect_type'],
    )



import factory

from flex.serializers.definitions import SwaggerDefinitionsSerializer
from flex.serializers.core import SwaggerSerializer
from flex.constants import EMPTY
from flex.http import (
    Request,
    Response,
)


class RequestFactory(factory.Factory):
    url = 'http://www.example.com/'
    method = 'get'
    content_type = 'application/json'
    body = EMPTY
    request = None
    headers = factory.Dict({})

    class Meta:
        model = Request


class ResponseFactory(factory.Factory):
    url = 'http://www.example.com/'
    content_type = 'application/json'
    content = EMPTY
    status_code = 200
    headers = factory.Dict({})

    request = factory.SubFactory(
        RequestFactory, url=factory.SelfAttribute('..url'),
    )

    class Meta:
        model = Response


def SchemaFactory(**kwargs):
    kwargs.setdefault('swagger', '2.0')
    kwargs.setdefault('info', {'title': 'Test API', 'version': '0.0.1'})
    kwargs.setdefault('paths', {})


    definitions_serializer = SwaggerDefinitionsSerializer(
        data=kwargs,
    )
    assert definitions_serializer.is_valid(), definitions_serializer.errors

    swagger_serializer = SwaggerSerializer(
        definitions_serializer.object,
        data=kwargs,
        context=definitions_serializer.object,
    )

    assert swagger_serializer.is_valid(), swagger_serializer.errors
    return swagger_serializer.object



import pytest

from flex.serializers.definitions import (
    SchemaSerializer,
    ItemsSerializer,
)

from flex.constants import (
    ARRAY,
    INTEGER,
    STRING,
)

from tests.utils import assert_error_message_equal


def test_empty_schema_is_valid():
    schema = {}

    serializer = SchemaSerializer(data=schema)

    assert serializer.is_valid()


def test_schema_reference_is_placed_in_deferred_refrences():
    context = {'deferred_references': set()}
    serializer = SchemaSerializer(
        data={'$ref': 'SomeReference'},
        context=context,
    )
    assert serializer.is_valid()
    assert 'SomeReference' in serializer.context['deferred_references']


def test_schema_item_references_are_deferred():
    context = {'deferred_references': set()}
    serializer = SchemaSerializer(
        data={'items': ['SomeReference']},
        context=context,
    )
    assert serializer.is_valid()
    assert 'SomeReference' in serializer.context['deferred_references']


#
# items validation
#
@pytest.mark.parametrize(
    'items',
    (
        1234,  # integer
        1.234,  # number
        True,  # boolean
        None,  # null
    ),
)
def test_items_invalid_when_not_array_or_object_or_reference(items):
    schema = {
        'type': ARRAY,
        'items': [items],
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert_error_message_equal(
        serializer.errors['items'][0],
        ItemsSerializer.default_error_messages['invalid_type_for_items'],
    )


def test_items_detects_invalid_single_schema():
    schema = {
        'type': ARRAY,
        'items': {
            'type': INTEGER,
            'minLength': 4,  # invalid with type 'integer'
        },
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert 'minLength' in serializer.errors['items'][0]
    assert_error_message_equal(
        serializer.errors['items'][0]['minLength'][0],
        serializer.error_messages['invalid_type_for_min_length'],
    )


def test_items_with_valid_singular_schema():
    schema = {
        'type': ARRAY,
        'items': {
            'type': INTEGER,
            'minimum': 4,
        },
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert 'items' not in serializer.errors


def test_items_detects_invalid_schema_in_array():
    schema = {
        'type': ARRAY,
        'items': [
            {
                'type': INTEGER,
                'minLength': 4,  # invalid with type 'integer'
            },
            {
                'type': STRING,
                'minimum': 4,  # invalid with type 'string'
            },
        ]
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert 'minLength' in serializer.errors['items'][0]
    assert_error_message_equal(
        serializer.errors['items'][0]['minLength'][0],
        serializer.error_messages['invalid_type_for_min_length'],
    )
    assert 'minimum' in serializer.errors['items'][1]
    assert_error_message_equal(
        serializer.errors['items'][1]['minimum'][0],
        serializer.error_messages['invalid_type_for_minimum'],
    )


def test_items_with_array_of_valid_schemas():
    schema = {
        'type': ARRAY,
        'items': [
            {
                'type': INTEGER,
                'minimum': 4,
            },
            {
                'type': STRING,
                'minLength': 4,
            },
        ]
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert 'items' not in serializer.errors


def test_items_with_mixed_array_of_references_and_schemas():
    schema = {
        'type': ARRAY,
        'items': [
            {
                'type': INTEGER,
                'minimum': 4,
            },
            'SomeReference',
            {
                'type': STRING,
                'minLength': 4,
            },
        ]
    }

    serializer = SchemaSerializer(
        data=schema,
        context={'deferred_references': set()}
    )

    assert 'items' not in serializer.errors



from __future__ import unicode_literals

import yaml

from flex.serializers.core import SwaggerSerializer
from flex.serializers.definitions import SwaggerDefinitionsSerializer


def generate(path='flex/schema.yaml'):
    with open(path) as stream:
        schema = yaml.load(stream)

    swagger_definitions_serializer = SwaggerDefinitionsSerializer(
        data=schema,
        context={'foo': 'bar'},
    )
    assert swagger_definitions_serializer.is_valid(), swagger_definitions_serializer.errors

    swagger_definitions = swagger_definitions_serializer.object

    swagger_serializer = SwaggerSerializer(
        data=schema,
        context=swagger_definitions,
    )

    assert swagger_serializer.is_valid(), swagger_serializer.errors

    return swagger_serializer



import functools

from django.core.exceptions import ValidationError

from flex.utils import chain_reduce_partial
from flex.context_managers import ErrorCollection
from flex.paths import (
    match_request_path_to_api_path,
)
from flex.validation.operation import (
    construct_operation_validators,
    validate_operation,
)
from flex.error_messages import MESSAGES
from flex.constants import REQUEST_METHODS
from flex.http import normalize_request


def validate_request_to_path(request, paths, base_path, context):
    """
    Given a request, check whether the path of the request matches any if the
    api paths.  Note that this does not do deep validation on the path
    parameters themselves, but only matches whether the request path *looks*
    like an api path.

    If so, return the api path and the path definitions.
    """
    try:
        api_path = match_request_path_to_api_path(
            path_definitions=paths,
            request_path=request.path,
            base_path=base_path,
        )
    except LookupError:
        raise ValidationError(MESSAGES['request']['unknown_path'])

    path_definition = paths[api_path] or {}
    return api_path, path_definition


def validate_request_method_to_operation(request, path_definition):
    """
    Given a request, validate that the request method is valid for the request
    path.

    If so, return the operation related to this request method.
    """
    method = request.method
    try:
        operation = path_definition[method]
    except KeyError:
        allowed_methods = set(REQUEST_METHODS).intersection(path_definition.keys())
        raise ValidationError(
            MESSAGES['request']['invalid_method'].format(
                method, allowed_methods,
            ),
        )
    return operation


def validate_request(request, paths, base_path, context, inner=False):
    """
    Request validation does the following steps.

       1. validate that the path matches one of the defined paths in the schema.
       2. validate that the request method conforms to a supported methods for the given path.
       3. validate that the request parameters conform to the parameter
          definitions for the operation definition.
    """
    with ErrorCollection(inner=inner) as errors:
        # 1
        try:
            api_path, path_definition = validate_request_to_path(
                request=request,
                paths=paths,
                base_path=base_path,
                context=context,
            )
        except ValidationError as err:
            errors['path'].extend(list(err.messages))
            return  # this causes an exception to be raised since errors is no longer falsy.

        if not path_definition:
            # TODO: is it valid to not have a definition for a path?
            return

        # 2
        try:
            operation_definition = validate_request_method_to_operation(
                request=request,
                path_definition=path_definition,
            )
        except ValidationError as err:
            errors['method'].append(err.message)
            return

        if operation_definition is None:
            # TODO: is this compliant with swagger, can path operations have a null
            # definition?
            return

        # 3
        operation_validators = construct_operation_validators(
            api_path=api_path,
            path_definition=path_definition,
            operation_definition=operation_definition,
            context=context,
        )
        try:
            validate_operation(request, operation_validators, inner=True)
        except ValidationError as err:
            errors['method'].append(err.messages)

    return operation_definition


def generate_request_validator(schema, **kwargs):
    request_validator = functools.partial(
        validate_request,
        paths=schema['paths'],
        base_path=schema.get('basePath', ''),
        context=schema,
        **kwargs
    )
    return chain_reduce_partial(
        normalize_request,
        request_validator,
    )



import urllib
import urllib2

import requests

from flex.http import (
    normalize_request,
)


#
#  Test normalizatin of the request object from the requests library
#
def test_request_normalization(httpbin):
    raw_response = requests.post(httpbin.url + '/post')

    request = normalize_request(raw_response.request)

    assert request.path == '/post'
    assert request.content_type is None
    assert request.url == httpbin.url + '/post'
    assert request.method == 'post'


def test_request_normalization_with_content_type(httpbin):
    raw_response = requests.post(
        httpbin.url + '/post',
        headers={'Content-Type': 'application/json'},
    )

    request = normalize_request(raw_response.request)

    assert request.path == '/post'
    assert request.content_type == 'application/json'
    assert request.url == httpbin.url + '/post'
    assert request.method == 'post'


#
# Test urllib request object
#
def test_urllib_request_normalization(httpbin):
    raw_request = urllib2.Request(
        httpbin.url + '/get',
        headers={'Content-Type': 'application/json'},
    )

    request = normalize_request(raw_request)

    assert request.path == '/get'
    assert request.content_type == 'application/json'
    assert request.url == httpbin.url + '/get'
    assert request.method == 'get'



from __future__ import unicode_literals

import urlparse
import os
import collections
import requests

import six
import json
import yaml

from django.core.exceptions import ValidationError

from flex.context_managers import ErrorCollection
from flex.serializers.core import (
    SwaggerSerializer,
    SchemaSerializer,
)
from flex.serializers.definitions import SwaggerDefinitionsSerializer
from flex.utils import prettify_errors
from flex.validation.request import generate_request_validator
from flex.validation.response import generate_response_validator


def load_source(source):
    """
    Common entry point for loading some form of raw swagger schema.

    Supports:
        - python object (dictionary-like)
        - path to yaml file
        - path to json file
        - file object (json or yaml).
        - json string.
        - yaml string.
    """
    if isinstance(source, collections.Mapping):
        return source

    elif hasattr(source, 'read') and callable(source.read):
        raw_source = source.read()
    elif os.path.exists(os.path.expanduser(str(source))):
        with open(os.path.expanduser(str(source)), 'r') as source_file:
            raw_source = source_file.read()
    elif isinstance(source, six.string_types):
        parts = urlparse.urlparse(source)
        if parts.scheme and parts.netloc:
            raw_source = requests.get(source).content
        else:
            raw_source = source

    try:
        try:
            return json.loads(raw_source)
        except ValueError:
            pass

        try:
            return yaml.load(raw_source)
        except yaml.scanner.ScannerError:
            pass
    except NameError:
        pass

    raise ValueError(
        "Unable to parse `{0}`.  Tried yaml and json.".format(source),
    )


def parse(raw_schema):
    definitions_serializer = SwaggerDefinitionsSerializer(
        data=raw_schema,
    )
    if not definitions_serializer.is_valid():

        message = "Swagger definitions did not validate:\n\n"
        message += prettify_errors(definitions_serializer.errors)
        raise ValueError(message)

    swagger_definitions = definitions_serializer.object

    swagger_serializer = SwaggerSerializer(
        swagger_definitions,
        data=raw_schema,
        context=swagger_definitions,
    )

    if not swagger_serializer.is_valid():
        message = "Swagger schema did not validate:\n\n"
        message += prettify_errors(swagger_serializer.errors)
        raise ValueError(message)

    return swagger_serializer.object


def load(target):
    """
    Given one of the supported target formats, load a swagger schema into it's
    python representation.
    """
    raw_schema = load_source(target)
    return parse(raw_schema)


def validate(schema, target=None, **kwargs):
    """
    Given the python representation of a JSONschema as defined in the swagger
    spec, validate that the schema complies to spec.  If `target` is provided,
    that target will be validated against the provided schema.
    """
    schema_serializer = SchemaSerializer(data=schema, **kwargs)
    if not schema_serializer.is_valid():
        message = "JSON Schema did not validate:\n\n"
        message += prettify_errors(schema_serializer.errors)
        raise ValueError(message)

    if target is not None:
        validator = schema_serializer.save()
        validator(target)


def validate_api_call(schema, request, response):
    with ErrorCollection() as errors:
        try:
            operation_definition = generate_request_validator(schema, inner=True)(request)
        except ValidationError as err:
            errors['request'].append(err.messages)
            return

        try:
            generate_response_validator(operation_definition, schema, inner=True)(response)
        except ValidationError as err:
            errors['response'].append(err.messages)



from __future__ import unicode_literals

import numbers
import collections
import six


SCHEMES = (
    'http', 'https', 'ws', 'wss',
)


MIMETYPES = (
    'application/json',
)


NULL = 'null'
BOOLEAN = 'boolean'
INTEGER = 'integer'
NUMBER = 'number'
STRING = 'string'
ARRAY = 'array'
OBJECT = 'object'

PRIMATIVE_TYPES = {
    '': (type(None),),
    None: (type(None),),
    NULL: (type(None),),
    BOOLEAN: (bool,),
    INTEGER: (int, long),
    NUMBER: (numbers.Number,),
    STRING: six.string_types,
    ARRAY: (collections.Sequence,),
    OBJECT: (collections.Mapping,),
}

TRUE_VALUES = set(('true', 'True', '1'))
FALSE_VALUES = set(('false', 'False', '0', ''))

HEADER_TYPES = (
    STRING,
    INTEGER,
    NUMBER,
    BOOLEAN,
    ARRAY,
)


PATH = 'path'
BODY = 'body'
QUERY = 'query'
FORM_DATA = 'formData'
HEADER = 'header'
PARAMETER_IN_VALUES = (
    QUERY,
    HEADER,
    PATH,
    FORM_DATA,
    BODY,
)


CSV = 'csv'
MULTI = 'multi'
SSV = 'ssv'
TSV = 'tsv'
PIPES = 'pipes'

COLLECTION_FORMATS = (
    CSV,
    SSV,
    TSV,
    PIPES,
    MULTI,
)

DELIMETERS = {
    CSV: ',',
    SSV: ' ',
    TSV: '\t',
    PIPES: '|',
}


API_KEY = 'apiKey'
BASIC = 'basic'
OAUTH_2 = 'oath2'
SECURITY_TYPES = (
    API_KEY,
    BASIC,
    OAUTH_2,
)


QUERY = QUERY
HEADER = HEADER
SECURITY_API_KEY_LOCATIONS = (
    QUERY,
    HEADER,
)


IMPLICIT = 'implicit'
PASSWORD = 'password'
APPLICATION = 'application'
ACCESS_CODE = 'accessCode'
SECURITY_FLOWS = (
    IMPLICIT,
    PASSWORD,
    APPLICATION,
    ACCESS_CODE,
)


class Empty(object):
    def __cmp__(self, other):
        raise TypeError('Empty cannot be compared to other values')


"""
Sentinal empty value for use with distinguishing `None` from a key not
being present.
"""
EMPTY = Empty()

UUID = 'uuid'
DATETIME = 'date-time'
EMAIL = 'email'
INT32 = 'int32'
INT64 = 'int64'
URI = 'uri'


FORMATS = (
    ('integer', 'int32'),
    ('integer', 'int64'),
    ('number', 'float'),
    ('number', 'double'),
    ('string', 'byte'),
    ('string', 'date'),
    ('string', 'date-time'),
    ('string', 'email'),
    ('string', 'uri'),
)


# Request Methods
GET = 'get'
PUT = 'put'
POST = 'post'
DELETE = 'delete'
OPTIONS = 'options'
HEAD = 'head'
PATCH = 'patch'

REQUEST_METHODS = (GET, PUT, POST, DELETE, OPTIONS, HEAD, PATCH)



from flex.serializers.definitions import (
    HeaderSerializer,
)

from flex.constants import (
    ARRAY,
    INTEGER,
)


def test_schema_reference_is_placed_in_deferred_refrences():
    context = {'deferred_references': set()}
    serializer = HeaderSerializer(
        data={'type': INTEGER, 'schema': 'SomeReference'},
        context=context,
    )
    assert serializer.is_valid(), serializer.errors
    assert 'SomeReference' in serializer.context['deferred_references']


def test_schema_item_references_are_deferred():
    context = {'deferred_references': set()}
    serializer = HeaderSerializer(
        data={'type': ARRAY, 'items': ['SomeReference']},
        context=context,
    )
    assert serializer.is_valid(), serializer.errors
    assert 'SomeReference' in serializer.context['deferred_references']



from flex.utils import is_value_of_type


class TypedDefaultMixin(object):
    default_error_messages = {
        'default_is_incorrect_type': (
            "The value supplied for 'default' must match the specified type."
        ),
    }

    def validate_default_type(self, attrs, errors):
        if 'default' in attrs and 'type' in attrs:
            if not is_value_of_type(attrs['default'], attrs['type']):
                errors['default'].append(
                    self.error_messages['default_is_incorrect_type'],
                )



from flex.serializers.definitions import (
    DefinitionsSerializer,
)

from tests.utils import assert_error_message_equal


def test_reference_within_reference():
    serializer = DefinitionsSerializer(
        data={
            'Node': {
                'properties': {
                    'parent': {'$ref': 'Unknown'},
                },
            },
        },
        context={'deferred_references': set()},
    )
    assert not serializer.is_valid()
    assert 'non_field_errors' in serializer.errors
    assert_error_message_equal(
        serializer.errors['non_field_errors'][0],
        serializer.error_messages['unknown_references'],
    )



import os

import click
from click.testing import CliRunner

from flex.cli import main


DIR = os.path.dirname(os.path.abspath(__file__))


def test_flex_cli_schema_validation():
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            '-s',
            os.path.join(DIR, 'schemas/cli-test-valid-schema.yaml'),
        ]
    )

    assert result.exit_code == 0
    assert result.output == 'Validation passed\n'


def test_flex_cli_schema_validation_with_verbose_argument_name():
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            '--source',
            os.path.join(DIR, 'schemas/cli-test-valid-schema.yaml'),
        ]
    )

    assert result.exit_code == 0
    assert result.output == 'Validation passed\n'


def test_flex_cli_schema_validation_with_invalid_schema():
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            '--source',
            os.path.join(DIR, 'schemas/cli-test-invalid-schema.yaml'),
        ]
    )

    assert result.exit_code == 1
    assert "Error: Swagger schema did not validate:" in result.output
    assert "'swagger'" in result.output
    assert "2.1" in result.output



from flex.serializers.core import ParameterSerializer

from tests.utils import assert_error_message_equal


def test_unknown_parameter_reference():
    serializer = ParameterSerializer(
        data=['SomeReference'],
        context={},
    )

    assert not serializer.is_valid()
    assert 'non_field_errors' in serializer.errors[0]
    assert_error_message_equal(
        serializer.errors[0]['non_field_errors'][0],
        serializer.error_messages['unknown_reference']
    )


def test_valid_parameter_reference():
    serializer = ParameterSerializer(
        data=['SomeReference'],
        context={
            'parameters': {'SomeReference': {}},
        },
    )

    assert serializer.is_valid()



import collections

from django.core.exceptions import ValidationError

from flex.exceptions import SafeNestedValidationError
from flex.utils import (
    prettify_errors,
)


class ErrorCollection(object):
    def __init__(self, inner=False, message='Invalid'):
        self.inner = inner
        self.message = message
        self.errors = collections.defaultdict(list)

    def __enter__(self):
        return self.errors

    def __exit__(self, type_, value, traceback):
        if any((type_, value, traceback)):
            if issubclass(type_, ValidationError):
                self.errors = value.messages
            else:
                return False
        if self.errors:
            if self.inner:
                raise SafeNestedValidationError(dict(self.errors))
            else:
                raise ValueError(self.message + '\n' + prettify_errors(self.errors))



import pytest

from flex.serializers.common import BaseItemsSerializer
from flex.serializers.core import (
    SchemaSerializer,
)
from flex.constants import (
    INTEGER,
    ARRAY,
    STRING,
)

from tests.utils import assert_error_message_equal

#
# $ref validation tests
#
def test_unknown_schema_reference():
    schema = {
        '$ref': 'SomeReference',
    }
    serializer = SchemaSerializer(
        data=schema,
        context={},
    )

    assert not serializer.is_valid()
    assert '$ref' in serializer.errors
    assert_error_message_equal(
        serializer.errors['$ref'][0],
        serializer.error_messages['unknown_reference']
    )


def test_valid_reference():
    schema = {
        '$ref': 'SomeReference',
    }
    serializer = SchemaSerializer(
        data=schema,
        context={
            'definitions': {'SomeReference': {}},
        },
    )

    assert serializer.is_valid()


#
# items validation tests
#
@pytest.mark.parametrize(
    'items',
    (
        1234,  # integer
        1.234,  # number
        True,  # boolean
        None,  # null
    ),
)
def test_items_invalid_when_not_array_or_object_or_reference(items):
    schema = {
        'type': ARRAY,
        'items': [items],
    }

    serializer = SchemaSerializer(
        data=schema,
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert_error_message_equal(
        serializer.errors['items'][0],
        BaseItemsSerializer.default_error_messages['invalid_type_for_items'],
    )


def test_items_as_missing_reference():
    schema = {
        'type': ARRAY,
        'items': 'SomeReference',
    }
    serializer = SchemaSerializer(
        data=schema,
    )
    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert_error_message_equal(
        serializer.errors['items'][0],
        serializer.error_messages['unknown_reference'],
    )


def test_items_as_existing_reference():
    schema = {
        'type': ARRAY,
        'items': 'SomeReference',
    }
    serializer = SchemaSerializer(
        data=schema,
        context={
            'definitions': {'SomeReference': {}},
        },
    )
    assert 'items' not in serializer.errors


def test_items_validated_as_valid_schema_object():
    schema = {
        'type': ARRAY,
        'items': {
            'type': 'unknown-type',
        }
    }
    serializer = SchemaSerializer(
        data=schema,
    )
    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert 'type' in serializer.errors['items'][0]


def test_items_as_array_of_invalid_schemas():
    schema = {
        'items': [
            {'type': 'unknown-type'},
            {'minLength': -1},
        ],
    }
    serializer = SchemaSerializer(
        data=schema,
    )
    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert 'type' in serializer.errors['items'][0]
    assert 'minLength' in serializer.errors['items'][1]


def test_items_as_array_of_valid_schemas():
    schema = {
        'type': ARRAY,
        'items': [
            {'type': INTEGER},
            {'minLength': 3},
        ],
    }
    serializer = SchemaSerializer(
        data=schema,
    )
    assert 'items' not in serializer.errors


def test_items_as_array_of_references_with_missing_reference():
    schema = {
        'type': ARRAY,
        'items': ['SomeReference', 'SomeOtherReference']
    }
    serializer = SchemaSerializer(
        data=schema,
        context={
            'definitions': {'SomeReference': {}},
        },
    )
    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert 'SomeOtherReference' in serializer.errors['items'][0]
    assert_error_message_equal(
        serializer.errors['items'][0],
        serializer.error_messages['unknown_reference'],
    )


def test_items_as_array_of_valid_references():
    schema = {
        'type': ARRAY,
        'items': ['SomeReference', 'SomeOtherReference']
    }
    serializer = SchemaSerializer(
        data=schema,
        context={
            'definitions': {
                'SomeReference': {},
                'SomeOtherReference': {},
            },
        },
    )
    assert 'items' not in serializer.errors


def test_items_with_mixed_array_of_references_and_schemas():
    schema = {
        'type': ARRAY,
        'items': [
            {
                'type': INTEGER,
                'minimum': 4,
            },
            'SomeReference',
            {
                'type': STRING,
                'minLength': 4,
            },
        ]
    }

    serializer = SchemaSerializer(
        data=schema,
        context={'definitions': {'SomeReference': {}}}
    )

    assert 'items' not in serializer.errors



from flex.serializers.core import PathsSerializer
from flex.error_messages import MESSAGES
from flex.constants import (
    PATH,
    INTEGER,
)

from tests.utils import assert_error_message_equal

def test_paths_serializers_preserves_empty_paths():
    """
    Ensure that paths that are defined bare without any additional information
    are preserved in the paths serializer.
    """
    paths = {
        '/get': None,
        '/post': {'responses': None},
    }
    serializer = PathsSerializer(data=paths)

    assert serializer.is_valid()
    assert '/get' in serializer.object
    assert '/post' in serializer.object


def test_path_serializer_allows_parameters_that_are_not_defined():
    paths = {
        '/get/{id}/': None,
    }
    serializer = PathsSerializer(data=paths)

    assert serializer.is_valid()


def test_path_serializer_enforces_all_path_parameters_to_be_in_api_path():
    paths = {
        '/get/no-parameters/': {
            'parameters': [
                {
                    'name': 'id',
                    'in': PATH,
                    'description': 'id',
                    'type': INTEGER,
                    'required': True,
                },
            ],
        }
    }
    serializer = PathsSerializer(data=paths)

    assert not serializer.is_valid()
    assert 'non_field_errors' in serializer.errors
    assert '/get/no-parameters/' in serializer.errors['non_field_errors'][0]
    assert_error_message_equal(
        serializer.errors['non_field_errors'][0]['/get/no-parameters/'][0],
        MESSAGES['path']['missing_parameter'],
    )


def test_path_serializer_path_parameter_validation_handles_references():
    paths = {
        '/get/no-parameters/': {
            'parameters': [
                'Id',
            ],
        }
    }
    context = {
        'parameters': {
            'Id': {
                'name': 'id',
                'in': PATH,
                'description': 'id',
                'type': INTEGER,
                'required': True,
            },
        }
    }
    serializer = PathsSerializer(data=paths, context=context)

    assert not serializer.is_valid()
    assert 'non_field_errors' in serializer.errors
    assert '/get/no-parameters/' in serializer.errors['non_field_errors'][0]
    assert_error_message_equal(
        serializer.errors['non_field_errors'][0]['/get/no-parameters/'][0],
        MESSAGES['path']['missing_parameter'],
    )



import re
import six


def assert_error_message_equal(formatted_msg, unformatted_msg):
    """
    Helper assertion for testing that a formatted error message matches the
    expected unformatted version of that error.
    """
    # Replace all `{}` style substitutions with `.*` so that we can run a regex
    # on the overall message, ignoring any pieces that would have been
    # dynamically inserted.
    if not isinstance(formatted_msg, six.string_types):
        raise ValueError(
            "formatted_msg must be a string: got `{0}`".format(
                repr(formatted_msg)),
        )
    if not isinstance(unformatted_msg, six.string_types):
        raise ValueError(
            "unformatted_msg must be a string: got `{0}`".format(
                repr(unformatted_msg)),
        )
    # replace any string formatters
    pattern = re.sub('\{.*\}', '.*', unformatted_msg)
    # replace any parenthesis
    pattern = re.sub('\(', '\(', pattern)
    pattern = re.sub('\)', '\)', pattern)

    if not re.compile(pattern).search(formatted_msg):
        raise AssertionError(
            "`{0}` not found in `{1}`".format(
                formatted_msg, unformatted_msg,
            )
        )


def generate_validator_from_schema(schema, **kwargs):
    from flex.serializers.core import SchemaSerializer

    serializer = SchemaSerializer(data=schema, **kwargs)
    assert serializer.is_valid(), serializer.errors

    validator = serializer.save()
    return validator



import pytest

from flex.utils import is_non_string_iterable
from flex.serializers.fields import (
    MaybeListCharField,
    SecurityRequirementReferenceField,
)

from django.core.exceptions import ValidationError

from rest_framework import serializers

from tests.utils import assert_error_message_equal


#
# MaybeListCharField tests
#
def test_maybe_list_char_field_accepts_strings():
    field = MaybeListCharField()
    data = {'foo': 'a-string'}
    into = {}
    field.field_from_native(data, {}, 'foo', into)
    assert data['foo'] == into.get('foo')


def test_maybe_list_char_field_accepts_lists():
    field = MaybeListCharField()
    data = {'foo': ['a-string', 'another-string']}
    into = {}
    field.field_from_native(data, {}, 'foo', into)
    assert data['foo'] == into.get('foo')


def test_maybe_list_char_field_runs_validators_on_singular_strings():
    def validator(value):
        if is_non_string_iterable(value):
            if not all([v.startswith('bar') for v in value]):
                raise ValidationError('error')
        else:
            if not value.startswith('bar'):
                raise ValidationError('error')

    field = MaybeListCharField(validators=[validator])
    data = {'foo': 'not-bar'}
    into = {}
    with pytest.raises(ValidationError):
        field.field_from_native(data, {}, 'foo', into)


def test_maybe_list_char_field_runs_validators_on_lists():
    def validator(value):
        if is_non_string_iterable(value):
            if not all([v.startswith('bar') for v in value]):
                raise ValidationError('error')
        else:
            if not value.startswith('bar'):
                raise ValidationError('error')

    field = MaybeListCharField(validators=[validator])
    data = {'foo': ['a-string', 'another-string']}
    into = {}
    with pytest.raises(ValidationError):
        field.field_from_native(data, {}, 'foo', into)


#
# SecurityRequirementReferenceField tests
#
def test_invalid_with_unknown_reference():
    class TestSerializer(serializers.Serializer):
        foo = SecurityRequirementReferenceField()

    serializer = TestSerializer(
        data={'foo': 'SomeReference'},
        context={
            'securityDefinitions': {},
        },
    )

    assert not serializer.is_valid()
    assert 'foo' in serializer.errors
    assert_error_message_equal(
        serializer.errors['foo'][0],
        SecurityRequirementReferenceField.default_error_messages['unknown_reference'],
    )


def test_valid_with_known_reference():
    class TestSerializer(serializers.Serializer):
        foo = SecurityRequirementReferenceField()

    serializer = TestSerializer(
        data={'foo': 'SomeReference'},
        context={
            'securityDefinitions': {'SomeReference': {}},
        },
    )

    assert serializer.is_valid()



import functools

from django.core.exceptions import ValidationError

from flex.utils import is_non_string_iterable
from flex.context_managers import ErrorCollection
from flex.validation.common import (
    generate_type_validator,
    generate_format_validator,
    generate_required_validator,
    generate_multiple_of_validator,
    generate_minimum_validator,
    generate_maximum_validator,
    generate_min_length_validator,
    generate_max_length_validator,
    generate_min_items_validator,
    generate_max_items_validator,
    generate_unique_items_validator,
    generate_pattern_validator,
    generate_enum_validator,
    validate_object,
    generate_value_processor,
)
from flex.validation.schema import (
    construct_schema_validators,
    generate_items_validator,
)
from flex.parameters import find_parameter
from flex.paths import path_to_regex
from flex.constants import EMPTY


def type_cast_parameters(parameter_values, parameter_definitions, context):
    typed_parameters = {}
    for key in parameter_values.keys():
        try:
            parameter_definition = find_parameter(parameter_definitions, name=key)
        except KeyError:
            continue
        value = parameter_values[key]
        value_processor = generate_value_processor(context=context, **parameter_definition)
        typed_parameters[key] = value_processor(value)
    return typed_parameters


def get_path_parameter_values(request_path, api_path, path_parameters, context):
    raw_values = path_to_regex(
        api_path,
        path_parameters,
    ).match(request_path).groupdict()
    return type_cast_parameters(raw_values, path_parameters, context=context)


def validate_path_parameters(request_path, api_path, path_parameters, context, inner=False):
    """
    Helper function for validating a request path
    """
    parameter_values = get_path_parameter_values(
        request_path, api_path, path_parameters, context,
    )
    validate_parameters(parameter_values, path_parameters, context=context, inner=inner)


def validate_query_parameters(raw_query_data, query_parameters, context, inner=False):
    query_data = {}
    for key, value in raw_query_data.items():
        if is_non_string_iterable(value) and len(value) == 1:
            query_data[key] = value[0]
        else:
            query_data[key] = value
    validate_parameters(query_data, query_parameters, context, inner=inner)


def validate_parameters(parameter_values, parameters, context, inner=False):
    validators = construct_multi_parameter_validators(parameters, context=context)

    with ErrorCollection(inner=inner) as errors:
        # we should have a validator for every parameter value
        assert not set(parameter_values.keys()).difference(validators.keys())

        for key, validator in validators.items():
            try:
                validator(parameter_values.get(key, EMPTY))
            except ValidationError as err:
                errors[key].extend(list(err.messages))


def construct_parameter_validators(parameter, context):
    """
    Constructs a dictionary of validator functions for the provided parameter
    definition.
    """
    validators = {}
    if 'schema' in parameter:
        validators.update(construct_schema_validators(parameter['schema'], context=context))
    for key in parameter:
        if key in validator_mapping:
            validators[key] = validator_mapping[key](context=context, **parameter)
    return validators


validator_mapping = {
    'type': generate_type_validator,
    'format': generate_format_validator,
    'required': generate_required_validator,
    'multipleOf': generate_multiple_of_validator,
    'minimum': generate_minimum_validator,
    'maximum': generate_maximum_validator,
    'minLength': generate_min_length_validator,
    'maxLength': generate_max_length_validator,
    'minItems': generate_min_items_validator,
    'maxItems': generate_max_items_validator,
    'uniqueItems': generate_unique_items_validator,
    'enum': generate_enum_validator,
    'pattern': generate_pattern_validator,
    'items': generate_items_validator,
}


def construct_multi_parameter_validators(parameters, context):
    """
    Given an iterable of parameters, returns a dictionary of validator
    functions for each parameter.  Note that this expects the parameters to be
    unique in their name value, and throws an error if this is not the case.
    """
    validators = {}
    for parameter in parameters:
        key = parameter['name']
        if key in validators:
            raise ValueError("Duplicate parameter name {0}".format(key))
        parameter_validators = construct_parameter_validators(parameter, context=context)
        validators[key] = functools.partial(
            validate_object,
            validators=parameter_validators,
            inner=True,
        )

    return validators



from __future__ import unicode_literals

import re
import urlparse
import operator

import six

from django.core.validators import RegexValidator

from rest_framework import serializers

from flex.formats import registry
from flex.decorators import maybe_iterable
from flex.utils import is_value_of_type
from flex.constants import (
    SCHEMES,
    PRIMATIVE_TYPES,
    PARAMETER_IN_VALUES,
    COLLECTION_FORMATS,
    HEADER_TYPES,
    SECURITY_TYPES,
    SECURITY_API_KEY_LOCATIONS,
    SECURITY_FLOWS,
    ARRAY,
)
from flex.error_messages import MESSAGES


def host_validator(value):
    parts = urlparse.urlparse(value)
    host = parts.netloc or parts.path
    if value != host:
        raise serializers.ValidationError(
            "Invalid host: {0}, expected {1}".format(value, host),
        )


def path_validator(value):
    if not value.startswith('/'):
        raise serializers.ValidationError("Path must start with a '/'")
    parts = urlparse.urlparse(value)
    if value != parts.path:
        raise serializers.ValidationError("Invalid Path: {0}".format(value))


@maybe_iterable
def scheme_validator(value):
    if value not in SCHEMES:
        raise serializers.ValidationError("Unknown scheme: {0}".format(value))


# top-level type name / [ tree. ] subtype name [ +suffix ] [ ; parameters ]

TOP_LEVEL_TYPE_NAMES = set((
    'application',
    'audio',
    'example',
    'image',
    'message',
    'model',
    'multipart',
    'text',
    'video',
))


MIMETYPE_PATTERN = (
    '^'
    '(application|audio|example|image|message|model|multipart|text|video)'  # top-level type name
    '/'
    '(vnd(\.[-a-zA-Z0-9]+)*\.)?'  # vendor tree
    '([-a-zA-Z0-9]+)'  # media type
    '(\+(xml|json|ber|der|fastinfoset|wbxml|zip))?'
    '((; [-a-zA-Z0-9]+=(([-\.a-zA-Z0-9]+)|(("|\')[-\.a-zA-Z0-9]+("|\'))))+)?'  # parameters
    '$'
)


@maybe_iterable
def mimetype_validator(value):
    return RegexValidator(MIMETYPE_PATTERN)(value)


@maybe_iterable
def string_type_validator(value):
    if not isinstance(value, six.string_types):
        raise serializers.ValidationError("Must be a string")


def format_validator(value):
    if value not in registry.formats:
        # TODO: unknown formats are ok, but want to be sure we have all of the
        # common ones before removing this.
        raise serializers.ValidationError('Unknown format: {0}'.format(value))


@maybe_iterable
def type_validator(value):
    if value not in PRIMATIVE_TYPES:
        raise serializers.ValidationError('Unknown type: {0}'.format(value))


def header_type_validator(value):
    if value not in HEADER_TYPES:
        raise serializers.ValidationError(
            MESSAGES['type']['invalid_header_type'].format(value),
        )


def parameter_in_validator(value):
    if value not in PARAMETER_IN_VALUES:
        raise serializers.ValidationError(
            "Unknown value for in: `{0}`".format(value),
        )


def collection_format_validator(value):
    if value not in COLLECTION_FORMATS:
        raise serializers.ValidationError(
            "Unknown collectionFormat: `{0}`".format(value),
        )


def MinValueValidator(minimum, allow_minimum=False):
    def validator(value):
        if allow_minimum:
            op = operator.ge
            msg_txt = "greater than or equal to"
        else:
            op = operator.gt
            msg_txt = "greater than"
        if not op(value, minimum):
            raise serializers.ValidationError(
                "{0} is not {1} `{2}`".format(
                    value, msg_txt, minimum,
                ),
            )

    return validator


def MaxValueValidator(maximum, allow_maximum=False):
    def validator(value):
        if allow_maximum:
            op = operator.le
            msg_txt = "less than or equal to"
        else:
            op = operator.lt
            msg_txt = "less than"
        if not op(value, maximum):
            raise serializers.ValidationError(
                "{0} is not {1} `{2}`".format(
                    value, msg_txt, maximum,
                ),
            )

    return validator


def security_type_validator(value):
    if value not in SECURITY_TYPES:
        raise serializers.ValidationError(
            "Unknown security type: {0}".format(value),
        )


def security_api_key_location_validator(value):
    if value not in SECURITY_API_KEY_LOCATIONS:
        raise serializers.ValidationError(
            "Unknown api key location: {0}".format(value),
        )


def security_flow_validator(value):
    if value not in SECURITY_FLOWS:
        raise serializers.ValidationError(
            "Unknown security flow: {0}".format(value),
        )


def regex_validator(value):
    try:
        re.compile(value)
    except re.error as e:
        raise serializers.ValidationError(
            "Invalid Regex: {0}".format(e.message)
        )


def is_array_validator(value):
    if not is_value_of_type(value, ARRAY):
        raise serializers.ValidationError(
            "Must be an array",
        )



from flex.serializers.definitions import (
    ParameterDefinitionsSerializer,
)
from flex.constants import (
    PATH,
    INTEGER,
)


def test_invalid_list_of_references():
    data = {
        'Id': [{'name': 'id', 'in': PATH, 'description': 'id', 'type': INTEGER, 'required': True}],
    }
    serializer = ParameterDefinitionsSerializer(
        data=data,
    )
    assert not serializer.is_valid(), serializer.object



import urllib
import urllib2

import requests

from flex.http import (
    normalize_response,
)


#
#  Test normalizatin of the response object from the requests library
#
def test_response_normalization(httpbin):
    raw_response = requests.get(httpbin.url + '/get')

    response = normalize_response(raw_response)

    assert response.path == '/get'
    assert response.content_type == 'application/json'
    assert response.url == httpbin.url + '/get'
    assert response.status_code == 200


#
#  Test normalization of urllib response object
#
def test_urllib_response_normalization(httpbin):
    raw_response = urllib.urlopen(httpbin.url + '/get')

    response = normalize_response(raw_response)

    assert response.path == '/get'
    assert response.content_type == 'application/json'
    assert response.url == httpbin.url + '/get'
    assert response.status_code == 200


#
#  Test normalization of urllib2 response object
#
def test_urllib2_response_normalization(httpbin):
    raw_response = urllib2.urlopen(httpbin.url + '/get')

    response = normalize_response(raw_response)

    assert response.path == '/get'
    assert response.content_type == 'application/json'
    assert response.url == httpbin.url + '/get'
    assert response.status_code == 200



import functools
import operator

from django.core.exceptions import ValidationError

from flex.utils import chain_reduce_partial
from flex.context_managers import ErrorCollection
from flex.validation.common import validate_object
from flex.validation.schema import construct_schema_validators
from flex.error_messages import MESSAGES
from flex.constants import (
    EMPTY,
)
from flex.validation.header import (
    construct_header_validators,
)
from flex.http import Response


def validate_status_code_to_response_definition(response, operation_definition):
    """
    Given a response, validate that the response status code is in the accepted
    status codes defined by this endpoint.

    If so, return the response definition that corresponds to the status code.
    """
    status_code = response.status_code
    operation_responses = operation_definition['responses']
    try:
        response_definition = operation_responses[status_code]
    except KeyError:
        raise ValidationError(
            MESSAGES['response']['invalid_status_code'].format(
                status_code, operation_responses.keys(),
            ),
        )
    return response_definition


def generate_response_body_validator(schema, context, **kwargs):
    validators = construct_schema_validators(schema, context=context)
    return chain_reduce_partial(
        operator.attrgetter('data'),
        functools.partial(
            validate_object,
            validators=validators,
            inner=True,
        ),
    )


def generate_response_header_validator(headers, context, **kwargs):
    validators = {}
    for key, header_definition in headers.items():
        header_validator = functools.partial(
            validate_object,
            validators=construct_header_validators(header_definition, context=context),
            inner=True,
        )
        # Chain the individual header validation function with a methodcaller
        # that will fetch the header with
        # `response.headers.get(header_name, EMPTY)`
        # and then feed that into the validation function.
        validators[key] = chain_reduce_partial(
            operator.methodcaller('get', key, EMPTY),
            header_validator,
        )
    return chain_reduce_partial(
        operator.attrgetter('headers'),
        functools.partial(validate_object, validators=validators, inner=True),
    )


validator_mapping = {
    'schema': generate_response_body_validator,
    'headers': generate_response_header_validator,
}


def generate_response_validator(response_definition, context):
    validators = {}
    for key in validator_mapping:
        if key in response_definition:
            validators[key] = validator_mapping[key](context=context, **response_definition)

    return functools.partial(
        validate_object,
        validators=validators,
        inner=True,
    )


def validate_response_content_type(response, content_types):
    assert isinstance(response, Response)  # TODO: remove this sanity check
    if response.content_type not in content_types:
        raise ValidationError(
            'Invalid content type `{0}`.  Must be one of `{1}`.'.format(
                response.content_type, content_types,
            ),
        )


def generate_response_content_type_validator(produces, **kwargs):
    return functools.partial(
        validate_response_content_type,
        content_types=produces,
    )


def validate_response(response, operation_definition, context, inner=False):
    """
    Response validation involves the following steps.
       4. validate that the response status_code is in the allowed responses for
          the request method.
       5. validate that the response content validates against any provided
          schemas for the responses.
       6. headers, content-types, etc..., ???
    """
    with ErrorCollection(inner=inner) as errors:
        # 4
        try:
            response_definition = validate_status_code_to_response_definition(
                response=response,
                operation_definition=operation_definition,
            )
        except ValidationError as err:
            errors['status_code'].append(err.message)
        else:
            # 5
            response_validator = generate_response_validator(
                response_definition,
                context=context,
            )
            try:
                response_validator(response)
            except ValidationError as err:
                errors['body'].extend(err.messages)

        # TODO: this should be merged with `response_body_validator`.
        response_content_type_validator = generate_response_content_type_validator(
            # TODO: this is a messy way to default to the global produces.
            produces=operation_definition.get('produces', context.get('produces', [])),
        )
        try:
            response_content_type_validator(response)
        except ValidationError as err:
            errors['produces'].append(err.messages)

        # 6
        # TODO



import pytest

from flex.serializers.common import BaseParameterSerializer

from flex.constants import (
    PATH,
    BODY,
    QUERY,
    FORM_DATA,
    BOOLEAN,
    NUMBER,
    ARRAY,
    NULL,
    STRING,
    MULTI,
    HEADER,
)

from tests.utils import assert_error_message_equal


def test_parameter_in_path_but_missing_required():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': PATH}
    )

    assert not serializer.is_valid()
    assert 'required' in serializer.errors
    assert_error_message_equal(
        serializer.errors['required'][0],
        serializer.error_messages['path_parameters_are_required'],
    )


def test_parameter_in_path_but_declared_not_required():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': PATH, 'required': False}
    )

    assert not serializer.is_valid()
    assert 'required' in serializer.errors
    assert_error_message_equal(
        serializer.errors['required'][0],
        serializer.error_messages['path_parameters_are_required'],
    )


def test_parameter_in_path_with_required_truthy():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': PATH, 'required': True}
    )

    assert 'required' not in serializer.errors


def assert_parameter_in_body_with_no_schema():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': BODY}
    )

    assert not serializer.is_valid()
    assert 'schema' in serializer.errors
    assert_error_message_equal(
        serializer.errors['schema'][0],
        serializer.error_messages['schema_required'],
    )


def assert_parameter_in_body_with_no_type():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': BODY}
    )

    assert not serializer.is_valid()
    assert 'type' in serializer.errors
    assert_error_message_equal(
        serializer.errors['type'][0],
        serializer.error_messages['type_required'],
    )


@pytest.mark.parametrize(
    'in_',
    (
        HEADER,
        PATH,
    )
)
def test_invalid_in_value_for_multi_collection_format(in_):
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': in_, 'collectionFormat': MULTI}
    )

    assert not serializer.is_valid()
    assert 'collectionFormat' in serializer.errors
    assert_error_message_equal(
        serializer.errors['collectionFormat'][0],
        serializer.error_messages['collection_format_must_be_multi'],
    )


@pytest.mark.parametrize(
    'in_',
    (
        QUERY,
        FORM_DATA,
    )
)
def test_valid_in_values_for_multi_collection_format(in_):
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': in_, 'collectionFormat': MULTI}
    )

    assert 'collectionFormat' not in serializer.errors


@pytest.mark.parametrize(
    'type_,default',
    (
        (BOOLEAN, 'abc'),
        (BOOLEAN, 1),
        (NUMBER, True),
        (NUMBER, 'abc'),
        (ARRAY, 'abc'),
        (NULL, False),
        (STRING, 1),
        (STRING, None),
    )
)
def test_mistyped_parameter_default_boolean_to_string(type_, default):
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': QUERY, 'type': type_, 'default': default}
    )

    assert not serializer.is_valid()
    assert 'default' in serializer.errors
    assert_error_message_equal(
        serializer.errors['default'][0],
        serializer.error_messages['default_is_incorrect_type'],
    )


def test_items_required_if_type_is_array():
    serializer = BaseParameterSerializer(
        data={'name': 'test', 'in': QUERY, 'type': ARRAY}
    )

    assert not serializer.is_valid()
    assert 'items' in serializer.errors
    assert_error_message_equal(
        serializer.errors['items'][0],
        serializer.error_messages['items_required'],
    )



from flex.serializers.core import (
    HeaderSerializer,
)
from flex.constants import (
    OBJECT,
)
from flex.error_messages import MESSAGES

from tests.utils import assert_error_message_equal


def test_header_type_cannot_be_object():
    serializer = HeaderSerializer(
        data={'type': OBJECT},
    )

    assert not serializer.is_valid()
    assert 'type' in serializer.errors
    assert_error_message_equal(
        serializer.errors['type'][0],
        MESSAGES['type']['invalid_header_type'],
    )



from __future__ import unicode_literals
import pytest

import six

from flex.serializers.validators import (
    host_validator,
    path_validator,
    scheme_validator,
    mimetype_validator,
    string_type_validator,
    type_validator,
    format_validator,
    parameter_in_validator,
    collection_format_validator,
    MinValueValidator,
    MaxValueValidator,
    security_type_validator,
    security_api_key_location_validator,
    security_flow_validator,
    regex_validator,
)
from flex.constants import (
    SCHEMES,
    MIMETYPES,
    FORMATS,
    PARAMETER_IN_VALUES,
    COLLECTION_FORMATS,
    NULL,
    INTEGER,
    SECURITY_TYPES,
    SECURITY_API_KEY_LOCATIONS,
    SECURITY_FLOWS,
)

from rest_framework import serializers


#
# type_validator tests
#
def test_type_validator_accepts_lists_of_types():
    """
    Ensure that the `type` validator accepts an iterable of types.
    """
    type_validator([NULL, INTEGER])


def test_type_validator_accepts_single_type():
    """
    Ensure that the `type` validator accepts a singular type.
    """
    type_validator(INTEGER)


def test_invalid_singular_type():
    with pytest.raises(serializers.ValidationError):
        type_validator('not-a-real-type')


def test_invalid_type_in_iterable_of_types():
    with pytest.raises(serializers.ValidationError):
        value = [
            NULL,
            'not-a-real-type',
            INTEGER,
        ]
        type_validator(value)


#
# format_validator tests
#
def test_format_sanity_check():
    with pytest.raises(serializers.ValidationError):
        format_validator('not-a-real-format')


def test_valid_format():
    format_validator(FORMATS[0][1])


#
# string_type_validator tests
#
def test_raises_for_non_strings():
    """
    Mostly just sanity checking
    """
    with pytest.raises(serializers.ValidationError):
        string_type_validator(1)

    with pytest.raises(serializers.ValidationError):
        string_type_validator(None)


def test_accepts_iterables():
    string_type_validator(['test'])


def test_accepts_bytes():
    string_type_validator(six.binary_type('test'))


def test_accepts_unicode():
    string_type_validator(six.text_type('test'))


#
# mimetype_validator tests
#
@pytest.mark.parametrize(
    'mimetype',
    (
        'application/json',
        'image/svg+xml',
        'application/vnd.oasis.opendocument.text',
        'text/plain; charset=utf-8',
        'video/mp4',
        'video/mp4; codecs="avc1.640028"',
        "video/mp4; codecs='avc1.640028'",
        'application/xhtml+xml',
        'image/png',
        'application/vnd.ms-excel',
    )
)
def test_mimetype_validator_on_valid_mimetypes(mimetype):
    mimetype_validator(mimetype)


@pytest.mark.parametrize(
    'mimetype',
    (
        'invalidtypename/json',
        'noslash',
    )
)
def test_mimetype_validator_on_invalid_mimetypes(mimetype):
    with pytest.raises(serializers.ValidationError):
        mimetype_validator(mimetype)


#
# scheme_validator tests
#
def test_scheme_invalid_value():
    with pytest.raises(serializers.ValidationError):
        scheme_validator('not-a-real-scheme')


def test_scheme_singular_value():
    scheme_validator(SCHEMES[0])


def test_scheme_iterable_value():
    scheme_validator([SCHEMES[0]])


#
# path_validator tests
#
def test_no_leading_slash():
    """
    Must begin with a leading `/`
    """
    with pytest.raises(serializers.ValidationError):
        path_validator('no-leading/slash/')


def test_with_leading_slash():
    path_validator('/with-leading/slash/')


def test_with_non_path_component():
    """
    Invalid if it contains extra stuff that isn't part of the path.
    """
    with pytest.raises(serializers.ValidationError):
        path_validator('/no-leading/slash/?foo=3')


#
# host_validator tests
#
def test_invalid_with_scheme():
    with pytest.raises(serializers.ValidationError):
        host_validator('http://www.example.com')


def test_subdomain_host():
    host_validator('www.example.com')


def test_only_host():
    host_validator('example.com')


def test_valid_with_port():
    host_validator('example.com:8000')


#
# parameter_in_validator tests
#
def test_invalid_in_value():
    with pytest.raises(serializers.ValidationError):
        parameter_in_validator('not-a-valid-in-value')


def test_valid_in_value():
    parameter_in_validator(PARAMETER_IN_VALUES[0])


#
# collection_format_validator tests
#
def test_invalid_collection_format():
    with pytest.raises(serializers.ValidationError):
        collection_format_validator('not-a-valid-collection-format')


def test_valid_collection_format():
    collection_format_validator(COLLECTION_FORMATS[0])


#
# MinValueValidator tests
#

def test_inclusive_minimum():
    validator = MinValueValidator(5, allow_minimum=True)

    validator(5)
    validator(100)

    with pytest.raises(serializers.ValidationError):
        validator(4)


def test_exclusive_minimum():
    validator = MinValueValidator(5, allow_minimum=False)

    validator(6)
    validator(100)

    with pytest.raises(serializers.ValidationError):
        validator(5)


#
# MaxValueValidator tests
#
def test_inclusive_maximum():
    validator = MaxValueValidator(5, allow_maximum=True)

    validator(5)
    validator(0)

    with pytest.raises(serializers.ValidationError):
        validator(6)


def test_exclusive_maximum():
    validator = MaxValueValidator(5, allow_maximum=False)

    validator(4)
    validator(0)

    with pytest.raises(serializers.ValidationError):
        validator(5)


#
# security_type_validator tests
#
def test_unknown_security_type():
    with pytest.raises(serializers.ValidationError):
        security_type_validator('not-a-real-security-type')


def test_with_known_security_type():
    security_type_validator(SECURITY_TYPES[0])


#
# security_api_key_location_validator tests
#
def test_with_unkown_api_key_location():
    with pytest.raises(serializers.ValidationError):
        security_api_key_location_validator('not-a-real-security-api-key-location')


def test_with_known_api_key_location():
    security_api_key_location_validator(SECURITY_API_KEY_LOCATIONS[0])


#
# security_flow_validator tests
#
def test_with_unknown_flow():
    with pytest.raises(serializers.ValidationError):
        security_flow_validator('not-a-real-security-flow')


def test_with_known_flow():
    security_flow_validator(SECURITY_FLOWS[0])


#
# regex_validator
#
def test_with_invalid_regex():
    with pytest.raises(serializers.ValidationError):
        regex_validator('[abc')


def test_with_valid_regex():
    regex_validator('abc')



from rest_framework import serializers

from flex.utils import is_non_string_iterable


class MaybeListCharField(serializers.CharField):
    def from_native(self, value):
        if is_non_string_iterable(value):
            return value
        return super(MaybeListCharField, self).from_native(value)


class SecurityRequirementReferenceField(serializers.CharField):
    """
    Field that references a defined security scheme declared in the Security
    Definitions.
    """
    default_error_messages = {
        'unknown_reference': "Unknown Security Scheme reference `{0}`",
    }

    def validate(self, value):
        if value not in self.context.get('securityDefinitions', {}):
            raise serializers.ValidationError(
                self.error_messages['unknown_reference'].format(value)
            )



import functools

from flex.utils import (
    is_non_string_iterable,
    is_value_of_any_type,
)
from flex.constants import EMPTY


def maybe_iterable(func):
    @functools.wraps(func)
    def inner(value):
        if is_non_string_iterable(value):
            return map(func, value)
        else:
            return func(value)
    return inner


def skip_if_not_of_type(*types):
    def outer(func):
        @functools.wraps(func)
        def inner(value, *args, **kwargs):
            if value is EMPTY or is_value_of_any_type(value, types):
                return func(value, *args, **kwargs)
        return inner
    return outer


RESERVED_WORDS = (
    'in',
    'format',
    'type',
)


def rewrite_reserved_words(func):
    """
    Given a function whos kwargs need to contain a reserved word such as 'in',
    allow calling that function with the keyword as 'in_', such that function
    kwargs are rewritten to use the reserved word.
    """
    @functools.wraps(func)
    def inner(*args, **kwargs):
        for word in RESERVED_WORDS:
            key = "{0}_".format(word)
            if key in kwargs:
                kwargs[word] = kwargs.pop(key)
        return func(*args, **kwargs)
    return inner


def suffix_reserved_words(func):
    """
    Given a function that is called with a reseved word, rewrite the keyword
    with an underscore suffix.
    """
    @functools.wraps(func)
    def inner(*args, **kwargs):
        for word in RESERVED_WORDS:
            if word in kwargs:
                key = "{0}_".format(word)
                kwargs[key] = kwargs.pop(word)
        return func(*args, **kwargs)
    return inner



from rest_framework.serializers import NestedValidationError


class SafeNestedValidationError(NestedValidationError):
    def __repr__(self):
        return 'ValidationError({0})'.format(self.messages)



import re
import functools

import iso8601
import rfc3987
import validate_email

from django.core.exceptions import ValidationError

from flex.utils import is_value_of_any_type
from flex.constants import (
    STRING,
    INTEGER,
    UUID,
    DATETIME,
    INT32,
    INT64,
    EMAIL,
    URI,
)
from flex.error_messages import MESSAGES


class FormatRegistry(object):
    def __init__(self):
        self.formats = {}

    def register(self, format_name, *types):
        if format_name in self.formats:
            raise ValueError(
                "The format `{0}` is already registered".format(format_name),
            )

        def outer(func):
            @functools.wraps(func)
            def inner(value, *args, **kwargs):
                """
                Format validation should only be executed if the value is of
                one of the appropriate types for the format.

                http://json-schema.org/latest/json-schema-validation.html#anchor105
                """
                if is_value_of_any_type(value, types):
                    return func(value, *args, **kwargs)

            self.formats[format_name] = inner
            return inner
        return outer

    def __getitem__(self, key):
        return self.formats[key]

    def __contains__(self, key):
        return key in self.formats


registry = FormatRegistry()
register = registry.register


"""
('number', 'float'),
('number', 'double'),
('string', 'byte'),
('string', 'date'),
"""


@register(URI, STRING)
def uri_validator(value):
    parts = rfc3987.parse(value, rule='URI')
    if not parts['scheme'] or not parts['authority']:
        raise ValidationError(MESSAGES['format']['invalid_uri'].format(value))


def number_of_bits(n):
    try:
        return n.bit_length()
    except AttributeError:
        return len(bin(n)) - 1


@register(INT32, INTEGER)
def int32_validator(value):
    num_bits = number_of_bits(value)
    if num_bits > 32:
        raise ValidationError(MESSAGES['format']['too_many_bits'].format(
            value, num_bits, 32
        ))


@register(INT64, INTEGER)
def int64_validator(value):
    num_bits = number_of_bits(value)
    if num_bits > 64:
        raise ValidationError(MESSAGES['format']['too_many_bits'].format(
            value, num_bits, 64
        ))


@register(EMAIL, STRING)
def email_validator(value):
    if not validate_email.validate_email(value):
        raise ValidationError(MESSAGES['format']['invalid_email'].format(value))


@register(DATETIME, STRING)
def date_time_format_validator(value):
    try:
        iso8601.parse_date(value)
    except iso8601.ParseError:
        raise ValidationError(MESSAGES['format']['invalid_datetime'].format(value))


UUID_PATTERN = re.compile(
    '^'
    '[a-f0-9]{8}-'
    '[a-f0-9]{4}-'
    '[1345][a-f0-9]{3}-'
    '[a-f0-9]{4}'
    '-[a-f0-9]{12}'
    '$'
)


@register(UUID, STRING)
def uuid_format_validator(value):
    if not UUID_PATTERN.match(value):
        raise ValidationError(MESSAGES['format']['invalid_uuid'].format(value))



import pytest

from flex.serializers.definitions import SecuritySchemeSerializer
from flex.constants import (
    BASIC,
    API_KEY,
    OAUTH_2,
    QUERY,
    HEADER,
    IMPLICIT,
    PASSWORD,
    APPLICATION,
    ACCESS_CODE,
)

from tests.utils import assert_error_message_equal


#
# name field validations tests
#
def test_name_required_if_type_is_api_key():
    serializer = SecuritySchemeSerializer(
        data={'type': API_KEY},
    )

    assert not serializer.is_valid()
    assert 'name' in serializer.errors
    assert_error_message_equal(
        serializer.errors['name'][0],
        serializer.error_messages['name_required']
    )


def test_type_as_api_key_with_name_value():
    serializer = SecuritySchemeSerializer(
        data={'type': API_KEY, 'name': 'TestName'},
    )

    assert 'name' not in serializer.errors


@pytest.mark.parametrize(
    'type_',
    (
        BASIC,
        OAUTH_2,
    ),
)
def test_name_optional_for_non_api_key_types(type_):
    assert 'name' not in SecuritySchemeSerializer(
        data={'type': type_},
    ).errors


#
# in field validation tests
#
def test_in_is_required_if_type_is_api_key():
    serializer = SecuritySchemeSerializer(
        data={'type': API_KEY, 'name': 'TestName'},
    )

    assert not serializer.is_valid()
    assert 'in' in serializer.errors
    assert_error_message_equal(
        serializer.errors['in'][0],
        serializer.error_messages['in_required']
    )


@pytest.mark.parametrize(
    'in_',
    (
        QUERY,
        HEADER,
    ),
)
def test_type_as_api_key_with_in_value(in_):
    serializer = SecuritySchemeSerializer(
        data={'type': API_KEY, 'in': in_},
    )

    assert 'in' not in serializer.errors


@pytest.mark.parametrize(
    'type_',
    (
        BASIC,
        OAUTH_2,
    ),
)
def test_in_optional_for_non_api_key_types(type_):
    assert 'in' not in SecuritySchemeSerializer(
        data={'type': type_},
    ).errors


#
# flow validations tests
#
def test_flow_is_required_if_type_oath2():
    serializer = SecuritySchemeSerializer(
        data={'type': OAUTH_2},
    )

    assert not serializer.is_valid()
    assert 'flow' in serializer.errors
    assert_error_message_equal(
        serializer.errors['flow'][0],
        serializer.error_messages['flow_required']
    )


def test_type_as_oath2_with_flow_value():
    serializer = SecuritySchemeSerializer(
        data={'type': OAUTH_2, 'flow': IMPLICIT},
    )

    assert 'flow' not in serializer.errors


def test_flow_optional_for_non_oath_types():
    assert 'flow' not in SecuritySchemeSerializer(
        data={'type': BASIC},
    ).errors
    assert 'flow' not in SecuritySchemeSerializer(
        data={'type': API_KEY},
    ).errors


#
# authorizationUrl validation tests
#
@pytest.mark.parametrize(
    'flow',
    (IMPLICIT, ACCESS_CODE)
)
def test_authorization_url_required_if_type_is_oath2(flow):
    serializer = SecuritySchemeSerializer(
        data={'type': OAUTH_2, 'flow': flow},
    )

    assert not serializer.is_valid()
    assert 'authorizationUrl' in serializer.errors
    assert_error_message_equal(
        serializer.errors['authorizationUrl'][0],
        serializer.error_messages['authorization_url_required']
    )


@pytest.mark.parametrize(
    'flow',
    (PASSWORD, APPLICATION)
)
def test_type_as_oath2_with_authorization_url_value(flow):
    serializer = SecuritySchemeSerializer(
        data={
            'type': OAUTH_2,
            'authorizationUrl': 'http://www.example.com/login/',
            'flow': flow,
        },
    )

    assert 'authorizationUrl' not in serializer.errors


@pytest.mark.parametrize(
    'type_',
    (BASIC, API_KEY)
)
def test_authorization_url_optional_for_non_oath_types(type_):
    assert 'authorizationUrl' not in SecuritySchemeSerializer(
        data={'type': type_},
    ).errors


@pytest.mark.parametrize(
    'flow',
    (PASSWORD, APPLICATION)
)
def test_authorization_url_optional_for_invalid_flows(flow):
    assert 'authorizationUrl' not in SecuritySchemeSerializer(
        data={'type': OAUTH_2, 'flow': flow},
    ).errors


#
# tokenUrl field validation tests
#
@pytest.mark.parametrize(
    'flow',
    (PASSWORD, APPLICATION, ACCESS_CODE)
)
def test_token_url_required_if_type_is_oath2(flow):
    serializer = SecuritySchemeSerializer(
        data={'type': OAUTH_2, 'flow': flow},
    )

    assert not serializer.is_valid()
    assert 'tokenUrl' in serializer.errors
    assert_error_message_equal(
        serializer.errors['tokenUrl'][0],
        serializer.error_messages['token_url_required']
    )


@pytest.mark.parametrize(
    'flow',
    (PASSWORD, APPLICATION, ACCESS_CODE)
)
def test_type_as_oath2_with_token_url_value(flow):
    serializer = SecuritySchemeSerializer(
        data={
            'type': OAUTH_2,
            'flow': flow,
            'tokenUrl': 'http://www.example.com/login/',
        },
    )

    assert 'tokenUrl' not in serializer.errors


@pytest.mark.parametrize(
    'type_',
    (BASIC, API_KEY)
)
def test_token_url_optional_for_non_oath_types(type_):
    assert 'tokenUrl' not in SecuritySchemeSerializer(
        data={'type': type_},
    ).errors


@pytest.mark.parametrize(
    'flow',
    (IMPLICIT,)
)
def test_token_url_optional_for_invalid_flows(flow):
    assert 'tokenUrl' not in SecuritySchemeSerializer(
        data={'type': OAUTH_2, 'flow': flow},
    ).errors
